module AG1_Print where
import List
import UU_Pretty
import UU_Pretty_ext
import OldPretty
import UU_Maps
import AG1_Parser
import AG1_Sem

printall res filename = print_data_types res filename
                        >-< ( text " " >-< print_functions res)
                        

print_data_types (gram, _, _, copytext) filename
    =   text ("module "++ filename ++ " where") 
       >-< (if null copytext then empty else vlist (map text (reverse copytext)))
--       >-< vlist (map printdatatype gram)
          
printdatatype (nt, alts)
 = text "data" >#< text nt >#< (pp_block "=" " " "| "  (map printalt alts) 
                                >-< text "deriving Show"
                               )
 
printalt (constr, rhs)  =     text constr 
                          >#< hv       [ pp_sp >#< text  ntname 
                                       | (elname, (ntname, _)) <- rhs 
                                       , elname /= "LHS" && elname /= "LOCAL"
                                       ]

print_functions (gr, inh, syn, _) 
  = vlist [prf nt_name con rhs 
          | (nt_name, alts) <- (gr :: Gram), syn ??  nt_name /= [],(con, rhs) <- alts
          ]
    where nonterminals = map fst gr 
          prf nt con rhs 
           = let synattr             = syn ?? nt
                 inhattr             = inh ?? nt
                 synt_funcs          = snd (rhs ?? "LHS")
                 syntexpr            = [e | e@(name, expr) <- synt_funcs
                                          , name `elem`    synattr] -- ag may be in error
                 errorexpr           = [e | e@(name, expr) <- synt_funcs
                                          , not (name `elem` synattr)]
                 rhsnames            = [ name | name <- map fst rhs
                                              , name /= "LOCAL" && name /= "LHS"]
                 rhscalls            = [ e | e@(_, (rhsnt,_)) <- rhs
                                           , rhsnt `elem` nonterminals
                                           , (not.null) (syn ?? rhsnt)
                                       ]
                 localdefs           = head [locdefs  | (v, (_,locdefs)) <- rhs
                                                      , v == "LOCAL"]
                 localattrs          = map fst localdefs
                 iarg                = sort (inhattr `union` (localattrs `union` rhsnames))
                 implicitargs        = nub iarg
                 print_local (name, expr) = text name >#< text " = " >#< text expr
                 print_elem  (name, (rhsnt, inhfuncs)) 
                        = hv [pp_block "( " " )" " , "  [text (name++"_"++s)| s <- syn ?? rhsnt] 
                             ,  text " =" >#< text name
                                >#< hv [pp_parens.text.((inhfuncs, implicitargs) ???) $ iarhs
                                       | iarhs <- (inh ?? rhsnt)
                                       ]
                             ]                                                            
             in  if    (not.null $ errorexpr)
                 then        error (   "non-existing synthesized attributes " ++ show errorexpr 
                                    ++ " in "  ++ show synattr 
                                    ++ " for " ++ show nt)
                 else if iarg /=  implicitargs 
                 then        error ("duplicate local names in "++nt++": "++show iarg)
                 else  text ("sem_"++nt++"_"++con) 
                       >|< hv [ if null rhsnames then empty else hlist (map (text.(' ':)) rhsnames)
                              , if null inhattr  then empty else hlist (map (text.(' ':)) inhattr)
                              ]
                       >-< indent 2 (text "=" >|< 
                                     (if null rhscalls && null localdefs 
                                      then empty 
                                      else  pp_block "let{ " "}in " "; " 
                                              (map print_elem  rhscalls ++  map print_local localdefs)
                                     ) >#<  pp_block "(" ")" ", "  
                                              [text.((syntexpr, implicitargs) ???) $ a | a <- synattr]
                                    )
                                                                      
(((a, b):rest), alternative) ??? v 
     | a == v = b
     | otherwise = (rest, alternative) ??? v
([], alt) ??? v = if v `elem` alt 
                  then v 
                  else error ("Lookup error and no copy rule possible: " ++ show v ++ "from" ++ show alt)
